#
# Copyright (c) 2022-present Didier Malenfant <didier@malenfant.net>
#
# This file is part of TraktorBuddy.
#
# TraktorBuddy is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TraktorBuddy is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License along with TraktorBuddy. If not,
# see <https://www.gnu.org/licenses/>.
#

from traktor_nowplaying import Listener as TraktorListener
from typing import Callable

from .Collection import Collection
from .Track import Track


# -- Class
class Listener:
    """Interface for listening to what Traktor is playing."""

    def __init__(self, collection: Collection, callback: Callable[[Track], None]):
        """Constructor a Traktor collection and a method to call when new tracks are posted."""

        self._collection = collection
        self._listener = TraktorListener(port=8000, quiet=True, custom_callback=self._new_track_callback)
        self._track_key_to_index = {}
        self._callback = callback

        # -- Build our track name-artist to index map
        for index in range(self._collection.numberOfTracks()):
            track = self._collection.trackAtIndex(index)

            title = track.title()
            if title is None:
                continue

            artist = track.artist()
            if artist is None:
                continue

            key = f'{title}{artist}'

            if key not in self._track_key_to_index:
                self._track_key_to_index[key] = index

    def _new_track_callback(self, data) -> None:
        info = dict(data)
        title = info.get('title', '')
        artist = info.get('artist', '')

        if len(title) == 0 or len(artist) == 0:
            return

        track_string = f'{title}{artist}'
        track_index = self._track_key_to_index.get(track_string, None)
        if track_index is not None:
            self._callback(self._collection.trackAtIndex(track_index))

    def start(self) -> None:
        self._listener.start()



#
# Copyright (c) 2022-present Didier Malenfant <didier@malenfant.net>
#
# This file is part of TraktorBuddy.
#
# TraktorBuddy is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TraktorBuddy is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License along with TraktorBuddy. If not,
# see <https://www.gnu.org/licenses/>.
#

import xml.etree.ElementTree as ET

from typing import List

from .Track import Track
from .TrackList import TrackList


# -- Class
class Playlist:
    """Interface for Traktor playlists."""

    def __init__(self, track_list: TrackList, node_element: ET.Element):
        """Constructor from an XML entry element."""

        self._node_element: ET.Element = node_element
        self._track_list: TrackList = track_list
        self._tracks: List[Track] = None

    def name(self) -> str:
        return self._node_element.get('NAME')

    def tracks(self) -> List[Track]:
        if self._tracks is not None:
            return self._tracks

        self._tracks = []

        playlist: Playlist = self._node_element.find('PLAYLIST')
        if playlist is None:
            return self._tracks

        if playlist.get('TYPE') != 'LIST':
            return self._tracks

        for entry in playlist.findall('ENTRY'):
            primary_key: ET.Element = entry.find('PRIMARYKEY')
            if primary_key is None or primary_key.get('TYPE') != 'TRACK':
                continue

            key: str = primary_key.get('KEY')
            if key is None:
                continue

            track: Track = self._track_list.trackWithPlaylistKey(key)
            if track is not None:
                self._tracks.append(track)

        return self._tracks



#
# Copyright (c) 2022-present Didier Malenfant <didier@malenfant.net>
#
# This file is part of TraktorBuddy.
#
# TraktorBuddy is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TraktorBuddy is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License along with TraktorBuddy. If not,
# see <https://www.gnu.org/licenses/>.
#

from .Collection import Collection
from .Color import Color
from .Folder import Folder
from .Key import OpenNotation
from .Listener import Listener
from .Playlist import Playlist
from .Rating import Rating
from .Track import Track

__all__ = ['Collection', 'Color', 'Folder', 'Listener', 'OpenNotation', 'Playlist', 'Rating', 'Track']



#
# Copyright (c) 2022-present Didier Malenfant <didier@malenfant.net>
#
# This file is part of TraktorBuddy.
#
# TraktorBuddy is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TraktorBuddy is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License along with TraktorBuddy. If not,
# see <https://www.gnu.org/licenses/>.
#

from enum import IntEnum, unique


@unique
class OpenNotation(IntEnum):
    Key_1d = 0    # -- C
    Key_8d = 1    # -- CsDb
    Key_3d = 2    # -- D
    Key_10d = 3   # -- DsEb
    Key_5d = 4    # -- E
    Key_12d = 5   # -- F
    Key_7d = 6    # -- FsGb
    Key_2d = 7    # -- G
    Key_9d = 8    # -- GsAb
    Key_4d = 9    # -- A
    Key_11d = 10  # -- AsBb
    Key_6d = 11   # -- B
    Key_10m = 12  # -- Cm
    Key_5m = 13   # -- CsmDbm
    Key_12m = 14  # -- Dm
    Key_7m = 15   # -- DsmEbm
    Key_2m = 16   # -- Em
    Key_9m = 17   # -- Fm
    Key_4m = 18   # -- FsmGbm
    Key_11m = 19  # -- Gm
    Key_6m = 20   # -- GsmAbm
    Key_1m = 21   # -- Am
    Key_8m = 22   # -- AsmBbm
    Key_3m = 23   # -- Bm



#
# Copyright (c) 2022-present Didier Malenfant <didier@malenfant.net>
#
# This file is part of
#
# TraktorBuddy is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TraktorBuddy is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License along with  If not,
# see <https://www.gnu.org/licenses/>.
#

import getopt
import sys
import os
import traceback
import time

from typing import List
from mutagen import File as MutagenFile
from pathlib import Path

from .__about__ import __version__
from .Exceptions import ArgumentError
from .Collection import Collection
from .Track import Track
from .Utility import Utility
from .Listener import Listener


_test_mode: bool = False
_verbose_mode: bool = False


def fix(_commands: List[str]) -> None:
    switch = {
        'labels': [fixLabels, 2, 3],
        'itunes': [fixItunes, 2, 3],
        'coverart': [fixCoverArt, 2, 3],
        'covercache': [fixCoverCache, 2, 2]
    }

    sub_command = _commands[1]
    method = switch.get(sub_command)
    if method is None:
        raise ArgumentError('Unknown argument \'' + sub_command + '\' to \'fix\' command.')

    nb_of_commands: int = len(_commands)
    if nb_of_commands < method[1]:
        raise ArgumentError('Expected an argument to \'fix\' command.')
    elif nb_of_commands > method[2]:
        raise ArgumentError('Too many arguments to \'fix\' command.')

    collection = Collection()
    tracks = collection.tracks() if nb_of_commands == method[1] else collection.findAllTracksAtPath(_commands[method[1]])
    method[0](collection, tracks)


def fixLabels(collection: Collection, tracks: List[Track]) -> None:
    global _test_mode
    global _verbose_mode

    nb_of_tracks_checked: int = 0
    nb_of_tracks_with_no_labels: int = 0
    nb_of_tracks_fixed: int = 0

    for track in tracks:
        if track.isASample():
            continue

        nb_of_tracks_checked += 1

        if track.label() is not None:
            continue

        nb_of_tracks_with_no_labels += 1

        file = MutagenFile(track.location())
        if file.tags is None:
            continue

        grouping = file.tags.get('©grp')
        if grouping is not None:
            if len(grouping) > 0:
                if _test_mode is False:
                    track.setLabel(grouping[0])

                nb_of_tracks_fixed += 1
        else:
            printed_location = False

            for tag in file.tags:
                if tag.startswith('APIC') or tag.startswith('PRIV:') or tag.startswith('GEOB:') or tag.startswith('POPM:') or tag.startswith('covr'):
                    continue
                if tag.startswith('©too'):
                    continue
                if tag.startswith('disk'):
                    continue
                if tag.startswith('©day'):
                    continue
                if tag.startswith('©nam'):
                    continue
                if tag.startswith('©alb'):
                    continue
                if tag.startswith('©ART'):
                    continue
                if tag.startswith('©gen'):
                    continue
                if tag.startswith('tmpo'):
                    continue
                if tag.startswith('----:com.apple.iTunes:initialkey'):
                    continue
                if tag.startswith('----:com.apple.iTunes:rating wmp'):
                    continue
                if tag.startswith('cpil'):
                    continue
                if tag.startswith('©cmt'):
                    continue
                if tag.startswith('TSSE'):
                    continue
                if tag.startswith('TPE1'):
                    continue
                if tag.startswith('TCMP'):
                    continue
                if tag.startswith('TIT2'):
                    continue
                if tag.startswith('TCON'):
                    continue
                if tag.startswith('TKEY'):
                    continue
                if tag.startswith('COMM::eng'):
                    continue
                if tag.startswith('TDRC'):
                    continue
                if tag.startswith('TBPM'):
                    continue
                if tag.startswith('TDOR'):
                    continue
                if tag.startswith('TDRL'):
                    continue
                if tag.startswith('TPE4'):
                    continue
                if tag.startswith('TENC'):
                    continue
                if tag.startswith('©wrt'):
                    continue
                if tag.startswith('trkn'):
                    continue
                if tag.startswith('TLEN'):
                    continue
                if tag.startswith('TXXX'):
                    continue
                if tag.startswith('TALB'):
                    continue
                if tag.startswith('TXXX'):
                    continue
                if tag.startswith('TRCK'):
                    continue
                if tag.startswith('aART'):
                    continue
                if tag.startswith('RVA2:SeratoGain'):
                    continue
                if tag.startswith('COMM:ID3v1 Comment'):
                    continue
                if tag.startswith('TSOT'):
                    continue
                if tag.startswith('TSOA'):
                    continue
                if tag.startswith('----:com.apple.iTunes:energylevel'):
                    continue
                if tag.startswith('----:com.apple.iTunes:iTunNORM'):
                    continue
                if tag.startswith('----:com.apple.iTunes:iTunSMPB'):
                    continue
                if tag.startswith('apID'):
                    continue
                if tag.startswith('cnID'):
                    continue
                if tag.startswith('ownr'):
                    continue
                if tag.startswith('plID'):
                    continue
                if tag.startswith('purd'):
                    continue
                if tag.startswith('stik'):
                    continue
                if tag.startswith('----:com.apple.iTunes:iTunMOVI'):
                    continue
                if tag.startswith('sonm'):
                    continue
                if tag.startswith('COMM:iTunSMPB:eng'):
                    continue
                if tag.startswith('COMM:iTunNORM:eng'):
                    continue
                if tag.startswith('TPOS'):
                    continue
                if tag.startswith('atID'):
                    continue
                if tag.startswith('geID'):
                    continue
                if tag.startswith('soal'):
                    continue
                if tag.startswith('soar'):
                    continue
                if tag.startswith('soco'):
                    continue
                if tag.startswith('----:com.apple.iTunes:Encoding Params'):
                    continue
                if tag.startswith('UFID:http://www.jhutveckling.se'):
                    continue
                if tag.startswith('GRP1'):
                    continue
                if tag.startswith('TIT1'):
                    continue
                if tag.startswith('rtng'):
                    continue
                if tag.startswith('cprt'):
                    continue
                if tag.startswith('sfID'):
                    continue
                if tag.startswith('xid '):
                    continue
                if tag.startswith('cmID'):
                    continue
                if tag.startswith('pgap'):
                    continue
                if tag.startswith('----:com.apple.iTunes:iTunes_CDDB_TrackNumber'):
                    continue

                if _verbose_mode is True:
                    if printed_location is False:
                        printed_location = True
                        print(track.location())

                    print("%s: %s" % (tag, file.tags[tag]))

    if _test_mode:
        print('Checked %d tracks. %d with no labels. %d of them can be fixed.' % (nb_of_tracks_checked, nb_of_tracks_with_no_labels, nb_of_tracks_fixed))
    else:
        print('Checked %d tracks. %d with no labels. Fixed %d of them.' % (nb_of_tracks_checked, nb_of_tracks_with_no_labels, nb_of_tracks_fixed))

    if _test_mode is False and nb_of_tracks_fixed > 0:
        collection.save()


def fixItunes(collection: Collection, tracks: List[Track]) -> None:
    global _test_mode
    global _verbose_mode

    nb_of_tracks_checked: int = 0
    nb_of_tracks_with_itunes_info: int = 0

    for track in tracks:
        if track.isASample():
            continue

        nb_of_tracks_checked += 1

        if track.getElement('ITUNES') is not None:
            continue

        nb_of_tracks_with_itunes_info += 1

        if _verbose_mode is True:
            print(track.location())

        if _test_mode is False:
            track.removeElement('ITUNES')

    if _test_mode:
        print('Checked %d tracks. %d with iTunes info tags.' % (nb_of_tracks_checked, nb_of_tracks_with_itunes_info))
    else:
        print('Checked %d tracks. Removed iTunes info tags from %d of them.' % (nb_of_tracks_checked, nb_of_tracks_with_itunes_info))

    if _test_mode is False and nb_of_tracks_with_itunes_info > 0:
        collection.save()


def fixCoverArt(collection: Collection, tracks: List[Track]) -> None:
    global _test_mode
    global _verbose_mode

    nb_of_tracks_checked: int = 0
    nb_of_tracks_with_missing_coverart_in_cache: int = 0
    track_filepaths_to_touch: List[str] = []

    for track in tracks:
        if track.isASample():
            continue

        nb_of_tracks_checked += 1

        cache_image_path = track.coverArtCacheFile(collection.folderPath())
        if cache_image_path is not None and os.path.exists(cache_image_path):
            continue

        if cache_image_path is None:
            file_coverart_image = track.coverArtImageFromFile()
            if file_coverart_image is None:
                continue

            file_coverart_image.close()

        nb_of_tracks_with_missing_coverart_in_cache += 1

        track_file_path = track.location()
        if _verbose_mode is True:
            print(track_file_path)

        if os.path.exists(track_file_path):
            track_filepaths_to_touch.append(track_file_path)

        if _test_mode is False:
            if cache_image_path is not None:
                track.setValueInInfoElement('COVERARTID', None)

    if _test_mode:
        print('Checked %d tracks. %d with missing or invalid cache coverart.' % (nb_of_tracks_checked, nb_of_tracks_with_missing_coverart_in_cache))
    else:
        print('Checked %d tracks. Removed invalid coverart ID from %d of them.' % (nb_of_tracks_checked, nb_of_tracks_with_missing_coverart_in_cache))

    if _test_mode is False and nb_of_tracks_with_missing_coverart_in_cache > 0:
        collection.save()

        # -- We need Traktor to spot that the track file is newer than the date in the collection
        time.sleep(5)

        for path in track_filepaths_to_touch:
            Path(path).touch()


def fixCoverCache(collection: Collection, tracks: List[Track]) -> None:
    global _test_mode
    global _verbose_mode

    cache_folder_path = os.path.join(collection.folderPath(), 'CoverArt')
    cache_files_found = {}

    nb_of_cache_entries: int = 0
    nb_of_orphan_cache_entries: int = 0

    for track in tracks:
        cache_file = track.coverArtCacheFile(collection.folderPath())
        if cache_file is None:
            continue

        cache_files_found[str(Path(cache_file).relative_to(cache_folder_path))[:-3]] = True

        nb_of_cache_entries += 1

    for p in Path(cache_folder_path).rglob('*'):
        if not os.path.isdir(p):
            continue

        for p in Path(p).rglob('*'):
            if os.path.isdir(p):
                continue

            cache_file = str(p.relative_to(cache_folder_path))[:-3]
            if cache_file in cache_files_found:
                continue

            nb_of_orphan_cache_entries += 1

            if _verbose_mode is True:
                print(str(p.relative_to(cache_folder_path)))

            if _test_mode is False:
                os.remove(p)

    if _test_mode:
        print('Checked %d cache entries. %d of wich are orphans.' % (nb_of_cache_entries, nb_of_orphan_cache_entries))
    else:
        print('Checked %d tracks. Deleted %d orphans.' % (nb_of_cache_entries, nb_of_orphan_cache_entries))


def tag(_commands: List[str]) -> None:
    if len(_commands) < 2:
        raise ArgumentError('Expected an argument to \'tag\' command.')

    switch = {
        'add': addTag,
        'remove': removeTag,
        'rename': renameTag,
        'years': addYearTag
    }

    sub_command = _commands[1]
    method = switch.get(sub_command)
    if method is None:
        raise ArgumentError('Unknown argument \'' + sub_command + '\' to \'tag\' command.')

    method(_commands)


def addTag(_commands: List[str]) -> None:
    global _test_mode
    global _verbose_mode

    nb_of_commands: int = len(_commands)
    if nb_of_commands > 4:
        raise ArgumentError('Too many arguments to \'add\' command.')
    elif nb_of_commands < 3:
        raise ArgumentError('Expected name argument to \'add\' command.')

    tag_name: str = _commands[2]
    if tag_name.__contains__(' '):
        raise ArgumentError('Tag names should not contain spaces.')

    collection = Collection()
    tracks = collection.tracks() if nb_of_commands == 3 else collection.findAllTracksAtPath(_commands[3])

    nb_of_tracks_tagged: int = 0
    for track in tracks:
        if track.isASample():
            continue

        existing_value: str = track.comments2()
        if existing_value is None:
            existing_value = tag_name
        elif tag_name in existing_value.split(' '):
            continue
        else:
            existing_value += ' ' + tag_name

        track.setComments2(existing_value)
        nb_of_tracks_tagged += 1

        if _verbose_mode is True:
            print(track.location())

    print('Tagged ' + str(nb_of_tracks_tagged) + ' tracks.')
    if _test_mode is False and nb_of_tracks_tagged > 0:
        collection.save()


def removeTag(_commands: List[str]) -> None:
    global _test_mode
    global _verbose_mode

    nb_of_commands: int = len(_commands)
    if nb_of_commands > 4:
        raise ArgumentError('Too many arguments to \'remove\' command.')
    elif nb_of_commands < 3:
        raise ArgumentError('Expected name argument to \'remove\' command.')

    tag_name: str = _commands[2]
    if tag_name.__contains__(' '):
        raise ArgumentError('Tag names should not contain spaces.')

    collection = Collection()
    tracks = collection.tracks() if nb_of_commands == 3 else collection.findAllTracksAtPath(_commands[3])

    nb_of_tracks_tagged: int = 0
    for track in tracks:
        if track.isASample():
            continue

        existing_value: str = track.comments2()
        if existing_value is None:
            continue

        names = existing_value.split(' ')
        if tag_name not in names:
            continue

        names.remove(tag_name)
        track.setComments2(" ".join(names))
        nb_of_tracks_tagged += 1

        if _verbose_mode is True:
            print(track.location())

    print('Removed tag from ' + str(nb_of_tracks_tagged) + ' tracks.')
    if _test_mode is False and nb_of_tracks_tagged > 0:
        collection.save()


def renameTag(_commands: List[str]) -> None:
    global _test_mode
    global _verbose_mode

    nb_of_commands: int = len(_commands)
    if nb_of_commands > 5:
        raise ArgumentError('Too many arguments to \'rename\' command.')
    elif nb_of_commands < 4:
        raise ArgumentError('Expected old and new name arguments to \'rename\' command.')

    old_tag_name: str = _commands[2]
    if old_tag_name.__contains__(' '):
        raise ArgumentError('Tag names should not contain spaces.')

    new_tag_name: str = _commands[3]
    if new_tag_name.__contains__(' '):
        raise ArgumentError('Tag names should not contain spaces.')

    collection = Collection()
    tracks = collection.tracks() if nb_of_commands == 4 else collection.findAllTracksAtPath(_commands[4])

    nb_of_tracks_tagged: int = 0
    for track in tracks:
        if track.isASample():
            continue

        existing_value: str = track.comments2()
        if existing_value is None:
            continue

        names = existing_value.split(' ')
        if old_tag_name not in names:
            continue

        names.remove(old_tag_name)
        track.setComments2(" ".join(names) + ' ' + new_tag_name)
        nb_of_tracks_tagged += 1

        if _verbose_mode is True:
            print(track.location())

    print('Renamed tag in ' + str(nb_of_tracks_tagged) + ' tracks.')
    if _test_mode is False and nb_of_tracks_tagged > 0:
        collection.save()


def addYearTag(_commands: List[str]) -> None:
    global _test_mode
    global _verbose_mode

    nb_of_commands: int = len(_commands)
    if nb_of_commands > 4:
        raise ArgumentError('Too many arguments to \'years\' command.')
    elif nb_of_commands < 3:
        raise ArgumentError('Expected old and new name arguments to \'rename\' command.')

    collection = Collection()
    tracks = collection.tracks() if nb_of_commands == 3 else collection.findAllTracksAtPath(_commands[3])

    nb_of_tracks_tagged: int = 0
    for track in tracks:
        if track.isASample():
            continue

        release_date = track.releaseDate()
        if release_date is None:
            continue

        year: int = track.releaseDate().year
        if year == 0:
            continue

        tag_name: str = 'Year:' + str(year)

        existing_value: str = track.comments2()
        if existing_value is None:
            existing_value = tag_name
        elif tag_name in existing_value.split(' '):
            continue
        else:
            existing_value += ' ' + tag_name

        if _test_mode is True:
            track.setComments2(existing_value)

        nb_of_tracks_tagged += 1

        if _verbose_mode is True:
            print(track.location())

    print('Tagged ' + str(nb_of_tracks_tagged) + ' tracks.')
    if _test_mode is False and nb_of_tracks_tagged > 0:
        collection.save()


def purgeBackups(_commands: List[str]) -> None:
    global _test_mode

    Collection.purgeBackups(test_mode=_test_mode)


def _new_track_posted(track: Track) -> None:
    label = track.label()
    if label is None:
        print(f'Playing "{track.title()}" by {track.artist()}.')
    else:
        print(f'Playing "{track.title()}" by {track.artist()} [{label}].')


def listen(_commands: List[str]) -> None:
    nb_of_commands: int = len(_commands)
    if nb_of_commands > 2:
        raise ArgumentError('Too many arguments to \'listen\' command.')

    collection: Collection = Collection()

    print('Listening to Traktor... (press CTRL-C to quit)')
    listener = Listener(collection, _new_track_posted)
    listener.start()


def printUsage(_commands: List[str]) -> None:
    method = None

    if len(_commands) > 1:
        switch = {
            'topics': printTopics,
            'tag': printTagUsage,
            'fix': printFixUsage,
            'license': printLicense
        }

        method = switch.get(_commands[1])
        if method is None:
            raise ArgumentError('Error: Unknown topic \'' + _commands[1] + '\'.')

        method()
        return

    printVersion()
    print('')
    print('usage: tktbud <options> <command> <arguments> <path>')
    print('')
    print('The following commands are supported:')
    print('')
    print('   help <topic>       - Show a help message. topic is optional (use \'help topics\' for a list).')
    print('   version            - Print the current version.')
    print('   tag <arguments>    - Add or remove tags (use \'help tag\' for a list of arguments).')
    print('   purge              - Purge all collection backups apart from the most recent one.')
    print('   listen             - Listen to tracks playing on Traktor and print info about them.')
    print('')
    print('The following options are supported:')
    print('')
    print('   --test/-t          - Run in test mode. Affected tracks are printed out. No changes are saved.')
    print('   --debug/-d         - Enable extra debugging information.')
    print('   --verbose/-v       - Enable verbose mode (prints information on the tracks affected).')
    print('')
    print('If path is provided then the action is only applied to the track contained in the Playlist/Folder at that path. Paths are / separated, i.e. \'/Folder1/Folder2/Playlist\'. Use \'\\ \' for spaces.')
    print('')
    print('If no path is provided the action is applied to ALL tracks in the collection.')
    print('')
    print('TraktorBuddy is free software, type "tktbud help license" for license information.')


def printVersion() -> None:
    print('🎧 Traktor Buddy v' + __version__ + ' 🎧')


def printTopics() -> None:
    printVersion()
    print('')
    print('Usage:')
    print('   tktbud help tag     - List arguments accepted by the tag command.')
    print('   tktbud help fix     - List arguments accepted by the fix command.')
    print('   tktbud help license - Show the license for the app.')
    print('')


def printTagUsage() -> None:
    printVersion()
    print('')
    print('Usage:')
    print('   tktbud tag add <name> <path>          - Add a tag named \'name\' to all tracks in \'path\'.')
    print('   tktbud tag delete <name> <path>       - Delete a tag named \'name\' for all tracks in \'path\'.')
    print('   tktbud tag rename <old> <new> <path>  - Rename tags named \'old\' to \'new\' for tracks in \'path\'.')
    print('   tktbud tag years <path>               - Add track\'s release year as a tag (i.e. Year:2022).')
    print('')


def printFixUsage() -> None:
    printVersion()
    print('')
    print('Usage:')
    print('   tktbud fix labels <path>   - Grab missing record labels from the track file\'s grouping field for all tracks in \'path\'.')
    print('   tktbud fix itunes <path>   - Delete iTunes info, if present, for all tracks in \'path\'.')
    print('   tktbud fix coverart <path> - Force Traktor to reload the coverart if the cached file is missing for all tracks in \'path\'.')
    print('   tktbud fix covercache      - Remove any files in the coverart cache that is not used by any tracks in the collection.')
    print('')


def printLicense() -> None:
    printVersion()
    print('')
    print('GPL License Version 3')
    print('')
    print('Copyright (c) 2024-present Didier Malenfant <didier@malenfant.net>')
    print('')
    print('TraktorBuddy is free software: you can redistribute it and/or modify it under the terms of the GNU General')
    print('Public License as published by the Free Software Foundation, either version 3 of the License, or')
    print('(at your option) any later version.')
    print('')
    print('TraktorBuddy is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the')
    print('implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public')
    print('License for more details.')
    print('')
    print('You should have received a copy of the GNU General Public License along with Main. If not,')
    print('see <https://www.gnu.org/licenses/>.')
    print('')


def main() -> None:
    global _test_mode
    global _verbose_mode

    _debug_on: bool = False

    try:
        # -- Gather the arguments, remove the first argument (which is the script filename)
        opts, _commands = getopt.getopt(sys.argv[1:], 'htdv', ['help', 'test', 'debug', 'verbose'])

        for o, a in opts:
            if o in ('-t', '--test'):
                print('Running in test mode.')
                _test_mode = True
            elif o in ('-d', '--debug'):
                print('Enabling debugging information.')
                _debug_on = True
            elif o in ('-v', '--verbose'):
                print('Enabling verbose mode.')
                _verbose_mode = True

        if len(_commands) == 0:
            raise ArgumentError('Expected a command! Maybe start with `tktbud help`?')

        switch = {
            'help': printUsage,
            'version': printVersion,
            'tag': tag,
            'purge': purgeBackups,
            'fix': fix,
            'listen': listen
        }

        if _commands is None:
            raise ArgumentError('Expected a command! Maybe start with `tktbud help`?')

        command = _commands[0]
        method = switch.get(command)
        if method is None:
            raise ArgumentError('Unknown commanwd \'' + command + '\'.')

        if _test_mode is not True and command != 'help' and command != 'version' and command != 'listen' and Utility.processIsRunning('Traktor'):
            raise RuntimeError('Traktor seems to be running. It\'s not a good idea to make changes to the collection at this time.')

        method(_commands)

    except getopt.GetoptError:
        printUsage([])
    except Exception as e:
        if _debug_on:
            print(traceback.format_exc())
        else:
            print(e)

        sys.exit(1)
    except KeyboardInterrupt:
        print('Execution interrupted by user.')
        pass


if __name__ == '__main__':
    main()



#
# Copyright (c) 2022-present Didier Malenfant <didier@malenfant.net>
#
# This file is part of TraktorBuddy.
#
# TraktorBuddy is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TraktorBuddy is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License along with TraktorBuddy. If not,
# see <https://www.gnu.org/licenses/>.
#

import xml.etree.ElementTree as ET

from typing import List

from .TrackList import TrackList
from .Playlist import Playlist
from .Track import Track


# -- Class
class Folder:
    """Interface for Traktor folders."""

    def __init__(self, track_list: TrackList, node_element: ET.Element):
        """Constructor from an XML entry element."""

        self._node_element: ET.Element = node_element
        self._track_list: TrackList = track_list
        self._folders: List[Folder] = None
        self._playlists: List[Playlist] = None

    def name(self) -> str:
        return self._node_element.get('NAME')

    def find(self, names: List[str]):
        name: str = names[0]
        nb_of_names: int = len(names)

        for playlist in self.playlists():
            if playlist.name() == name:
                if nb_of_names == 1:
                    return playlist

        for folder in self.folders():
            if folder.name() == name:
                if nb_of_names == 1:
                    return folder
                else:
                    return folder.find(names[1:])

        return None

    def findFolder(self, names: List[str]):
        result = self.find(names)
        if type(result) != Folder:
            return None

        return result

    def findPlaylist(self, names: List[str]):
        result = self.find(names)
        if type(result) != Playlist:
            return None

        return result

    def folders(self) -> List['Folder']:
        if self._folders is not None:
            return self._folders

        self._folders = []

        subnodes: ET.Element = self._node_element.find('SUBNODES')
        if subnodes is None:
            return self._folders

        for node in subnodes.findall('NODE'):
            if node.get('TYPE') != 'FOLDER':
                continue

            self._folders.append(Folder(self._track_list, node))

        return self._folders

    def playlists(self) -> List[Playlist]:
        if self._playlists is not None:
            return self._playlists

        self._playlists = []

        subnodes: ET.Element = self._node_element.find('SUBNODES')
        if subnodes is None:
            return self._playlists

        for node in subnodes.findall('NODE'):
            if node.get('TYPE') != 'PLAYLIST':
                continue

            self._playlists.append(Playlist(self._track_list, node))

        return self._playlists

    def tracks(self) -> List[Track]:
        result: List[Track] = []

        for folder in self.folders():
            result = result + folder.tracks()

        for playlist in self.playlists():
            result = result + playlist.tracks()

        return result



#
# Copyright (c) 2022-present Didier Malenfant <didier@malenfant.net>
#
# This file is part of TraktorBuddy.
#
# TraktorBuddy is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TraktorBuddy is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License along with TraktorBuddy. If not,
# see <https://www.gnu.org/licenses/>.
#

import pytz
import subprocess
import os
import xml.etree.ElementTree as ET

from datetime import datetime
from typing import List


# -- Class
class Utility:
    """Helper methods."""

    # -- This is used in Unit tests to mock the time for 'now'.
    _mock_now_date = None

    @classmethod
    def stringToInt(cls, string: str) -> int:
        if string is None:
            return None

        return int(string)

    @classmethod
    def stringToFloat(cls, string: str) -> float:
        if string is None:
            return None

        return float(string)

    @classmethod
    def dateFromString(cls, string: str, format: str, utc: bool = False) -> datetime:
        if string is None:
            return None

        try:
            date = datetime.strptime(string, format)
            if utc:
                date = pytz.utc.localize(date)

            return date
        except ValueError:
            return None

    @classmethod
    def utcTimeNow(cls) -> datetime:
        if Utility._mock_now_date is not None:
            return Utility._mock_now_date

        return datetime.now().astimezone(pytz.utc)

    @classmethod
    def utcDatetime(cls, year: int, month: int, day: int, hour: int, minutes: int, seconds: int) -> datetime:
        return pytz.utc.localize(datetime(year, month, day, hour, minutes, seconds))

    @classmethod
    def xmlElementToString(cls, element: ET.Element, xml_declaration: bool = False) -> str:
        return ET.tostring(element, encoding='unicode', short_empty_elements=False, xml_declaration=xml_declaration)

    @classmethod
    def processIsRunning(cls, process_name: str) -> bool:
        return process_name in Utility.shellCommand(['ps', '-axc', '-o', 'comm'])

    @classmethod
    def shellCommand(cls, command_and_args: List[str], from_dir: str = None) -> str:
        try:
            if from_dir is None:
                from_dir = os.path.expanduser('~')

            process = subprocess.Popen(command_and_args, cwd=from_dir, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            stdout, stderr = process.communicate()

            if process.returncode != 0:
                print(command_and_args)
                print(stdout)
                print(stderr)

                raise RuntimeError('Error running shell command.')

            return stdout.decode('UTF-8').split('\n')
        except RuntimeError:
            raise
        except SyntaxError:
            raise
        except Exception as e:
            raise RuntimeError('Error running shell command: ' + str(e))



#
# Copyright (c) 2022-present Didier Malenfant <didier@malenfant.net>
#
# This file is part of TraktorBuddy.
#
# TraktorBuddy is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TraktorBuddy is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License along with TraktorBuddy. If not,
# see <https://www.gnu.org/licenses/>.
#

class ArgumentError(Exception):
    """Error caused when command line arguments have something wrong in them."""
    pass



#
# Copyright (c) 2022-present Didier Malenfant <didier@malenfant.net>
#
# This file is part of TraktorBuddy.
#
# TraktorBuddy is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TraktorBuddy is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License along with TraktorBuddy. If not,
# see <https://www.gnu.org/licenses/>.
#

import os
import xml.etree.ElementTree as ET

from datetime import datetime
from PIL import Image
from mutagen import File as MutagenFile
from io import BytesIO

from .Utility import Utility
from .Key import OpenNotation
from .Rating import Rating
from .Color import Color


# -- Class
class Track:
    """Interface for Traktor tracks."""

    def __init__(self, entry_element: ET.Element):
        """Constructor from an XML entry element."""

        self._entry_element: ET.Element = entry_element
        self._info_element: ET.Element = self._entry_element.find('INFO')
        self._album_element: ET.Element = self._entry_element.find('ALBUM')
        self._modified: bool = False

    def _entryElement(self) -> ET.Element:
        return self._entry_element

    def _getInfoElement(self) -> ET.Element:
        if self._info_element is None:
            self._info_element = ET.SubElement(self._entry_element, 'INFO')

        return self._info_element

    def _getAlbumElement(self) -> ET.Element:
        if self._album_element is None:
            self._album_element = ET.SubElement(self._entry_element, 'ALBUM')

        return self._album_element

    def _getFromAlbumElement(self, key: str) -> str:
        if self._album_element is None:
            return None

        return self._album_element.get(key)

    def _markAsModified(self) -> None:
        self._modified = True

    def _playlistKey(self) -> str:
        location = self._entry_element.find('LOCATION')

        if location is None:
            return None

        webaddress = location.get('WEBADDRESS')
        if webaddress is not None:
            return webaddress

        volume = location.get('VOLUME')

        if volume is None:
            return None

        directory = location.get('DIR')

        if directory is None:
            return None

        file = location.get('FILE')

        if file is None:
            return None

        return volume + directory + file

    def updateModificationDateIfModified(self, date: datetime) -> None:
        if self._modified is False:
            return

        self._entry_element.set('MODIFIED_DATE', date.strftime('%Y/%m/%d'))
        self._entry_element.set('MODIFIED_TIME', str(date.second + (date.minute * 60) + (date.hour * 3600)))

        self._modified = False

    def getElement(self, name: str) -> ET.Element:
        return self._entry_element.find(name) is not None

    def removeElement(self, name: str):
        element = self._entry_element.find(name)
        if element is not None:
            self._entry_element.remove(element)
            self._markAsModified()

    def getValueFromInfoElement(self, name: str) -> str:
        if self._info_element is None:
            return None

        return self._info_element.get(name)

    def setValueInInfoElement(self, name: str, value: str):
        if self._info_element is None:
            if value is None:
                return

            self._info_element = ET.SubElement(self._entry_element, 'INFO')

        if value is None:
            self._info_element.attrib.pop(name)
        else:
            self._info_element.set(name, value)

        self._markAsModified()

    def isASample(self) -> bool:
        return self._entry_element.find('LOOPINFO') is not None

    def location(self) -> str:
        playlist_key = self._playlistKey()
        if playlist_key is None:
            return None

        if playlist_key.startswith('beatport:'):
            return playlist_key

        return '/Volumes/' + playlist_key.replace('/:', '/')

    def modificationDate(self) -> datetime:
        date = Utility.dateFromString(self._entry_element.get('MODIFIED_DATE'), '%Y/%m/%d')
        if date is None:
            return None

        seconds = Utility.stringToInt(self._entry_element.get('MODIFIED_TIME'))
        if seconds is None:
            return date

        seconds = seconds % (24 * 3600)
        hour = seconds // 3600
        seconds %= 3600
        minutes = seconds // 60
        seconds %= 60

        # -- Traktor modification dates are stored in UTC time.
        return Utility.utcDatetime(date.year, date.month, date.day, hour, minutes, seconds)

    def title(self) -> str:
        return self._entry_element.get('TITLE')

    def setTitle(self, value) -> None:
        self._entry_element.set('TITLE', value)
        self._markAsModified()

    def artist(self) -> str:
        return self._entry_element.get('ARTIST')

    def setArtist(self, value) -> None:
        self._entry_element.set('ARTIST', value)
        self._markAsModified()

    def beatgridLocked(self) -> bool:
        return self._entry_element.get('LOCK') == '1'

    def setBeatGridLocked(self, value: bool) -> None:
        if value is True:
            string = '1'
        else:
            string = '0'

        self._entry_element.set('LOCK', string)
        self._markAsModified()

        date = Utility.utcTimeNow()
        self._entry_element.set('LOCK_MODIFICATION_TIME', date.strftime('%Y-%m-%dT%H:%M:%S'))

    def beatgridLockModifiedDate(self) -> datetime:
        string = self._entry_element.get('LOCK_MODIFICATION_TIME')
        if string is None:
            return None

        return Utility.dateFromString(string, '%Y-%m-%dT%H:%M:%S', utc=True)

    def bitrate(self) -> int:
        return Utility.stringToInt(self.getValueFromInfoElement('BITRATE'))

    def setBitrate(self, value: int) -> None:
        self.setValueInInfoElement('BITRATE', str(value))

    def genre(self) -> str:
        return self.getValueFromInfoElement('GENRE')

    def setGenre(self, value: str) -> None:
        self.setValueInInfoElement('GENRE', value)

    def label(self) -> str:
        return self.getValueFromInfoElement('LABEL')

    def setLabel(self, value: str) -> None:
        self.setValueInInfoElement('LABEL', value)

    def producer(self) -> str:
        return self.getValueFromInfoElement('PRODUCER')

    def setProducer(self, value: str) -> None:
        self.setValueInInfoElement('PRODUCER', value)

    def mix(self) -> str:
        return self.getValueFromInfoElement('MIX')

    def setMix(self, value: str) -> None:
        self.setValueInInfoElement('MIX', value)

    def release(self) -> str:
        return self._getFromAlbumElement('TITLE')

    def setRelease(self, value: str) -> None:
        self.setValueInInfoElement('TITLE', value)

    def trackNumber(self) -> int:
        return Utility.stringToInt(self._getFromAlbumElement('TRACK'))

    def setTrackNumber(self, value: int) -> None:
        self._getAlbumElement().set('TRACK', str(value))
        self._markAsModified()

    def comments(self) -> str:
        return self.getValueFromInfoElement('COMMENT')

    def setComments(self, value: str) -> None:
        self.setValueInInfoElement('COMMENT', value)

    def comments2(self) -> str:
        return self.getValueFromInfoElement('RATING')

    def setComments2(self, value: str) -> None:
        self.setValueInInfoElement('RATING', value)

    def remixer(self) -> str:
        return self.getValueFromInfoElement('REMIXER')

    def setRemixer(self, value: str) -> None:
        self.setValueInInfoElement('REMIXER', value)

    def key(self) -> str:
        return self.getValueFromInfoElement('KEY')

    def setKey(self, value: str) -> None:
        self.setValueInInfoElement('KEY', value)

    def playCount(self) -> int:
        return Utility.stringToInt(self.getValueFromInfoElement('PLAYCOUNT'))

    def setPlayCount(self, value: int) -> None:
        self.setValueInInfoElement('PLAYCOUNT', str(value))

    def length(self) -> float:
        return Utility.stringToFloat(self.getValueFromInfoElement('PLAYTIME_FLOAT'))

    def setLength(self, value: float) -> None:
        self.setValueInInfoElement('PLAYTIME', str(round(value)))
        self.setValueInInfoElement('PLAYTIME_FLOAT', '{:.06f}'.format(value))

    def rating(self) -> Rating:
        # -- The following works with rekordbox and Serato too:
        # --    Unrated -> 0, 1-51 -> 1, 52-102 -> 2, 103-153 -> 3, 154-204 -> 4, 205-anything -> 5
        value = Utility.stringToInt(self.getValueFromInfoElement('RANKING'))

        if value is None:
            return None

        if value == 0:
            return Rating.Unrated
        elif value < 52:
            return Rating.OneStar
        elif value < 103:
            return Rating.TwoStars
        elif value < 154:
            return Rating.ThreeStars
        elif value < 205:
            return Rating.FourStars
        elif value <= 255:
            return Rating.FiveStars

        return None

    def setRating(self, value: Rating) -> None:
        map = {
            Rating.Unrated: 0,
            Rating.OneStar: 51,
            Rating.TwoStars: 102,
            Rating.ThreeStars: 153,
            Rating.FourStars: 205,
            Rating.FiveStars: 255
        }

        self.setValueInInfoElement('RANKING', str(map[value]))

    def importDate(self) -> datetime:
        return Utility.dateFromString(self.getValueFromInfoElement('IMPORT_DATE'), '%Y/%m/%d')

    def setImportDate(self, value: datetime) -> None:
        self.setValueInInfoElement('IMPORT_DATE', value.strftime('%Y/%m/%d'))

    def lastPlayedDate(self) -> datetime:
        return Utility.dateFromString(self.getValueFromInfoElement('LAST_PLAYED'), '%Y/%m/%d')

    def setLastPlayedDate(self, value: datetime) -> None:
        self.setValueInInfoElement('LAST_PLAYED', value.strftime('%Y/%m/%d'))

    def releaseDate(self) -> datetime:
        return Utility.dateFromString(self.getValueFromInfoElement('RELEASE_DATE'), '%Y/%m/%d')

    def setReleaseDate(self, value: datetime) -> None:
        self.setValueInInfoElement('RELEASE_DATE', value.strftime('%Y/%m/%d'))

    def fileSize(self) -> int:
        return Utility.stringToInt(self.getValueFromInfoElement('FILESIZE'))

    def setFileSize(self, value: int) -> None:
        self.setValueInInfoElement('FILESIZE', str(value))

    def bpm(self) -> float:
        tempo_element = self._entry_element.find('TEMPO')
        if tempo_element is None:
            return None

        return Utility.stringToFloat(tempo_element.get('BPM'))

    def setBpm(self, value: float) -> None:
        tempo_element = self._entry_element.find('TEMPO')
        if tempo_element is None:
            tempo_element = ET.SubElement(self._entry_element, 'TEMPO')

        tempo_element.set('BPM', '{:.06f}'.format(value))
        tempo_element.set('BPM_QUALITY', '100.000000')
        self._markAsModified()

    def traktorKey(self) -> OpenNotation:
        key_element = self._entry_element.find('MUSICAL_KEY')
        if key_element is None:
            return None

        value = Utility.stringToInt(key_element.get('VALUE'))
        if value is None:
            return None

        try:
            result = OpenNotation(value)
        except ValueError:
            return None

        return result

    def setTraktorKey(self, value: OpenNotation) -> None:
        key_element = self._entry_element.find('MUSICAL_KEY')
        if key_element is None:
            key_element = ET.SubElement(self._entry_element, 'MUSICAL_KEY')

        key_element.set('VALUE', str(int(value)))
        self._markAsModified()

    def color(self) -> Color:
        value = Utility.stringToInt(self.getValueFromInfoElement('COLOR'))

        if value is None:
            return None

        try:
            result = Color(value)
        except ValueError:
            return None

        return result

    def setColor(self, value: Color) -> None:
        self.setValueInInfoElement('COLOR', str(int(value)))

    def coverArtImageFromFile(self) -> Image:
        track_file_path = self.location()
        if not os.path.exists(track_file_path):
            return None

        try:
            # -- Mutagen can automatically detect format and type of tags
            file = MutagenFile(track_file_path)

            # -- Access APIC frame and grab the image
            tag = file.tags.get('APIC:', None)

            artwork_data = None
            if tag is not None:
                artwork_data = BytesIO(tag.data)
            else:
                cover_list = file.get('covr', None)
                if cover_list is not None and len(cover_list) != 0:
                    # -- We only use the first cover from the list
                    artwork_data = BytesIO(cover_list[0])

            if artwork_data is None:
                return None

            return Image.open(artwork_data)
        except Exception:
            return None

        return None

    def coverArtCacheFile(self, collection_folder_path: str) -> str:
        cover_art_id = self.getValueFromInfoElement('COVERARTID')
        if cover_art_id is None:
            return None

        return os.path.join(collection_folder_path, 'CoverArt', cover_art_id) + '000'

    def coverArtImageFromCache(self, collection_folder_path: str) -> Image:
        database_image_path = self.coverArtCacheFile(collection_folder_path)
        if database_image_path is None or not os.path.exists(database_image_path):
            return None

        artwork_file = open(database_image_path, "rb")
        data = artwork_file.read()
        artwork_file.close()

        if data[0] != 8:
            return None

        width = ((data[4] << 24) | (data[3] << 16) | (data[2] << 8) | data[1])
        height = ((data[8] << 24) | (data[7] << 16) | (data[6] << 8) | data[5])

        rgba_data = bytearray()

        # -- Re-order the color components from little endian data.
        for pixel_index in range(0, width * height):
            data_index = 9 + (pixel_index * 4)

            rgba_data.append(data[data_index + 2])
            rgba_data.append(data[data_index + 1])
            rgba_data.append(data[data_index])
            rgba_data.append(data[data_index + 3])

        return Image.frombytes('RGBA', (width, height), bytes(rgba_data))



#
# Copyright (c) 2022-present Didier Malenfant <didier@malenfant.net>
#
# This file is part of TraktorBuddy.
#
# TraktorBuddy is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TraktorBuddy is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License along with TraktorBuddy. If not,
# see <https://www.gnu.org/licenses/>.
#

import os

import xml.etree.ElementTree as ET

from typing import List
from semver import VersionInfo
from time import sleep
from pathlib import Path

from .Utility import Utility
from .Track import Track
from .TrackList import TrackList
from .Folder import Folder
from .Exceptions import ArgumentError


# -- Classes
class Collection:
    """Interface for Traktor collection."""

    def __init__(self, collection_path: str = None, _mock_element: ET.Element = None):
        """Constructor from a collection path, or it will just use the latest collection if no path is provided."""

        self._nml_element = None

        if _mock_element is None:
            if collection_path is None:
                self._collection_path = Collection.traktorCollectionFilePath()

                if self._collection_path is None:
                    raise RuntimeError('Error: Could not find any Traktor folder in \'' + Collection.nativeInstrumentsFolderPath() + '\'.')
            else:
                self._collection_path = collection_path

            self._collection_folder_path = Path(self._collection_path).parent

            print('Parsing Traktor collection in \'' + self._collection_path + '\'.')

            self._nml_element = ET.ElementTree(file=self._collection_path).getroot()
        else:
            self._nml_element = _mock_element

        self._track_list = TrackList(self._nml_element.find('COLLECTION'))

    def folderPath(self) -> str:
        return self._collection_folder_path

    def makeBackup(self):
        # -- Backups filename have a timestamp so we make sure to wait so that names cannot clash.
        sleep(1)

        backup_folder = Collection.traktorCollectionBackupFolderPath()
        os.makedirs(backup_folder, exist_ok=True)

        arguments: List[str] = ['zip', '-j', Utility.utcTimeNow().strftime('%Y-%m-%d-%H-%M-%S.zip'), self._collection_path]
        Utility.shellCommand(arguments, backup_folder)

    def save(self):
        self.makeBackup()

        date = Utility.utcTimeNow()

        for track in self.tracks():
            track.updateModificationDateIfModified(date)

        with open(self._collection_path, 'w') as out_file:
            out_file.write(Utility.xmlElementToString(self._nml_element, xml_declaration=True))

        print('Saved Traktor collection in \'' + self._collection_path + '\'.')

    def findAllTracksAtPath(self, path: str) -> List[Track]:
        crate = self.rootFolder().find(path.split('/'))
        if crate is None:
            raise RuntimeError('Could not find any folder or playlist at \'' + path + '\'.')
        else:
            return crate.tracks()

    def tracks(self) -> List[Track]:
        return self._track_list.tracks()

    def numberOfTracks(self) -> int:
        return len(self._track_list.tracks())

    def trackAtIndex(self, index: int) -> Track:
        if index >= self.numberOfTracks():
            raise ArgumentError("Out of bound access to a track.")

        return self._track_list.tracks()[index]

    def rootFolder(self) -> Folder:
        playlists_element: ET.Element = self._nml_element.find('PLAYLISTS')
        if playlists_element is None:
            return None

        root_node: ET.Element = playlists_element.find('NODE')
        if root_node is None:
            return None

        return Folder(self._track_list, root_node)

    def trackWithPlaylistKey(self, key) -> Track:
        return self._track_list.trackWithPlaylistKey(key)

    @classmethod
    def purgeBackups(cls, test_mode: bool = False):
        backup_folder: str = Collection.traktorCollectionBackupFolderPath()
        backup_list: List[str] = os.listdir(backup_folder)

        nb_of_backups: int = len(backup_list)
        if nb_of_backups < 2:
            print('No backups to purge.')
            return

        if test_mode is False:
            backup_list.sort()

            for file in backup_list[:-1]:
                os.remove(os.path.join(backup_folder, file))

        print('Purged ' + str(nb_of_backups - 1) + ' backup(s).')

    @classmethod
    def traktorCollectionBackupFolderPath(cls) -> str:
        return os.path.join(Collection.latestTraktorFolderPath(), 'Backup', 'TraktorBuddy')

    @classmethod
    def nativeInstrumentsFolderPath(cls) -> str:
        return os.path.join(os.path.expanduser('~'), 'Documents', 'Native Instruments')

    @classmethod
    def latestTraktorFolderPath(cls) -> str:
        base_folder = Collection.nativeInstrumentsFolderPath()

        lastest_version = None

        for path in os.listdir(base_folder):
            if not path.startswith('Traktor '):
                continue

            try:
                version = VersionInfo.parse(path[8:])

                if lastest_version is None or version > lastest_version:
                    lastest_version = version
            except ValueError:
                continue

        if lastest_version is None:
            return None

        return os.path.join(base_folder, 'Traktor ' + str(lastest_version))

    @classmethod
    def traktorCollectionFilePath(cls) -> str:
        return os.path.join(Collection.latestTraktorFolderPath(), 'collection.nml')



#
# Copyright (c) 2022-present Didier Malenfant <didier@malenfant.net>
#
# This file is part of TraktorBuddy.
#
# TraktorBuddy is free software: you can redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# TraktorBuddy is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
# implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License along with TraktorBuddy. If not,
# see <https://www.gnu.org/licenses/>.
#

import xml.etree.ElementTree as ET

from typing import List

from .Track import Track


# -- Class
class TrackList:
    """Interface for Traktor track list inside a collection."""

    def __init__(self, collection_element: ET.Element):
        """Constructor from an XML collection element."""

        self._collection_element: ET.Element = collection_element
        self._tracks: List[Track] = None

    def tracks(self) -> List[Track]:
        if self._collection_element is None:
            return []

        if self._tracks is not None:
            return self._tracks

        self._tracks = []

        for entry in self._collection_element.findall('ENTRY'):
            track = Track(entry)
            if track.location() is not None:
                self._tracks.append(track)

        return self._tracks

    def trackWithPlaylistKey(self, key) -> Track:
        for track in self.tracks():
            if track._playlistKey() == key:
                return track

        return None
